import asyncio
from pathlib import Path
from typing import TYPE_CHECKING, Literal, cast

import nonebot
from nonebot_plugin_htmlrender.data_source import TEMPLATES_PATH, env, get_new_page, markdown, read_tpl

if TYPE_CHECKING:
    from nonebot.drivers.none import Driver as NoneDriver


async def md_to_pic(
    md: str = "",
    extra_css: str = "",
    type: Literal["jpeg", "png"] = "png",  # noqa: A002
    quality: int | None = None,
    device_scale_factor: float = 2,
) -> bytes:
    template = env.get_template("markdown.html")
    md = markdown.markdown(
        md,
        extensions=[
            "pymdownx.tasklist",
            "tables",
            "fenced_code",
            "codehilite",
            "mdx_math",
            "pymdownx.tilde",
        ],
        extension_configs={"mdx_math": {"enable_dollar_delimiter": True}},
        tab_length=2,
    )

    extra = ""
    if "math/tex" in md:
        katex_css, katex_js, mathtex_js = await asyncio.gather(
            read_tpl("katex/katex.min.b64_fonts.css"),
            read_tpl("katex/katex.min.js"),
            read_tpl("katex/mathtex-script-type.min.js"),
        )
        extra = (
            f'<style type="text/css">{katex_css}</style>'
            f"<script defer>{katex_js}</script>"
            f"<script defer>{mathtex_js}</script>"
        )

    github_md_css, pygments_css = await asyncio.gather(
        read_tpl("github-markdown-light.css"),
        read_tpl("pygments-default.css"),
    )
    css = f"{github_md_css}\n{pygments_css}\n{extra_css}"

    html = await template.render_async(css=css, md=md, extra=extra)

    async with get_new_page(device_scale_factor) as page:
        await page.goto(f"file://{TEMPLATES_PATH}")
        await page.set_content(html, wait_until="networkidle")
        await page.wait_for_load_state("load")
        elem = await page.query_selector("article.markdown-body")
        assert elem
        return await elem.screenshot(type=type, quality=quality)


driver = nonebot.get_driver()


@driver.on_startup
async def _():
    project_root = Path(__file__).parent.parent.parent
    help_md_path = project_root / "docs" / "usage.md"
    help_img_path = project_root / "YetAnotherPicSearch" / "res" / "usage.jpg"
    if not (p := help_img_path.parent).exists():
        p.mkdir(parents=True)

    help_md = help_md_path.read_text("u8")
    help_img = await md_to_pic(help_md, type="jpeg")
    help_img_path.write_bytes(help_img)

    cast("NoneDriver", driver).exit()


nonebot.run()



from typing import cast

from lxml.html import HTMLParser, fromstring
from pyquery import PyQuery


class NHentaiItem:
    def __init__(self, data: PyQuery):
        self.origin: PyQuery = data  # 原始数据
        self.title: str = cast("str", data.find(".caption").text())
        cover = data.find(".cover")
        self.href: str = cast("str", cover.attr("href"))
        self.thumbnail: str = cast("str", cover.find("img").attr("data-src"))
        self.type: str = ""
        self.date: str = ""
        self.tags: list[str] = []

    def with_base_url(self, base: str) -> str:
        return f"{base}/{self.href}"


class NHentaiResponse:
    def __init__(self, resp_text: str, resp_href: str):
        self.origin: str = resp_text  # 原始数据
        uft8_parser = HTMLParser(encoding="utf-8")
        data = PyQuery(fromstring(self.origin, parser=uft8_parser))
        self.raw: list[NHentaiItem] = [NHentaiItem(i) for i in data.find(".gallery").items()]
        self.href: str = resp_href

    def with_base_url(self, base: str) -> str:
        return f"{base}/{self.href}"



from typing import cast

import arrow
from httpx import AsyncClient
from lxml.html import HTMLParser, fromstring
from nonebot_plugin_alconna.uniseg import UniMessage
from pyquery import PyQuery

from ..config import config
from ..nhentai_model import NHentaiItem, NHentaiResponse
from ..utils import (
    combine_message,
    filter_results_with_ratio,
    handle_img,
    parse_cookies,
    preprocess_search_query,
    shorten_url,
)

NHENTAI_HEADERS = (
    {"User-Agent": config.nhentai_useragent} if config.nhentai_cookies and config.nhentai_useragent else None
)
NHENTAI_COOKIES = parse_cookies(config.nhentai_cookies)


def get_nh_display_base():
    return config.nhentai_base_url if config.hide_nhentai_base_url else "https://nhentai.net"


async def update_nhentai_info(item: NHentaiItem) -> None:
    async with AsyncClient(
        headers=NHENTAI_HEADERS,
        cookies=NHENTAI_COOKIES,
        proxy=config.proxy,
    ) as session:
        resp = await session.get(item.with_base_url(config.nhentai_base_url))
        uft8_parser = HTMLParser(encoding="utf-8")
        data = PyQuery(fromstring(resp.text, parser=uft8_parser))
        item.origin = data
        item.title = cast(
            "str",
            (data.find("h2.title").text() if data.find("h2.title") else data.find("h1.title").text()),
        )
        item.type = cast("str", data.find('#tags a[href^="/category/"] .name').text())
        item.date = cast("str", data.find("#tags time").attr("datetime"))
        item.tags = [cast("str", i.text()) for i in data.find('#tags a:not([href*="/search/?q=pages"]) .name').items()]


async def nhentai_title_search(title: str) -> list[UniMessage]:
    query = preprocess_search_query(title)
    async with AsyncClient(
        headers=NHENTAI_HEADERS,
        cookies=NHENTAI_COOKIES,
        proxy=config.proxy,
    ) as session:
        resp = await session.get(
            f"{config.nhentai_base_url}/search/",
            params={"q": query},
        )
        if res := NHentaiResponse(
            resp.text,
            str(resp.url).replace(f"{config.nhentai_base_url}/", "", 1),
        ):
            # 只保留标题和搜索关键词相关度较高的结果，并排序，以此来提高准确度
            if res.raw:
                res.raw = filter_results_with_ratio(res, title)
            return await search_result_filter(res)

        return [UniMessage.text("NHentai 暂时无法使用")]


async def search_result_filter(res: NHentaiResponse) -> list[UniMessage]:
    display_base = get_nh_display_base()
    url = await shorten_url(
        res.with_base_url(display_base),
        force_shorten=True,
    )
    if not res.raw:
        return [UniMessage.text(f"NHentai 搜索结果为空\n搜索页面：{url}")]

    for i in res.raw:
        await update_nhentai_info(i)

    # 优先找汉化版；没找到就优先找原版
    if chinese_res := [i for i in res.raw if "translated" in i.tags and "chinese" in i.tags]:
        selected_res = chinese_res[0]
    elif not_translated_res := [i for i in res.raw if "translated" not in i.tags]:
        selected_res = not_translated_res[0]
    else:
        selected_res = res.raw[0]

    thumbnail = await handle_img(selected_res.thumbnail)
    date = arrow.get(selected_res.date).to("Asia/Shanghai").format("YYYY-MM-DD HH:mm")
    res_list = [
        "NHentai 搜索结果",
        thumbnail,
        selected_res.title,
        f"类型：{selected_res.type}",
        f"日期：{date}",
        f"来源：{selected_res.with_base_url(display_base)}",
        f"搜索页面：{url}",
    ]
    return [combine_message(res_list)]



from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Optional, TypeAlias, TypeVar

from httpx import AsyncClient
from nonebot_plugin_alconna.uniseg import UniMessage

SearchFunctionReturnTuple: TypeAlias = tuple[
    list[UniMessage],
    Optional["SearchFunctionType"],
]
SearchFunctionReturnType: TypeAlias = list[UniMessage] | SearchFunctionReturnTuple
SearchFunctionType: TypeAlias = Callable[
    [bytes, AsyncClient, str],
    Awaitable[SearchFunctionReturnType],
]

TSF = TypeVar("TSF", bound=SearchFunctionType)


@dataclass
class SearchFunctionInfo:
    func: SearchFunctionType


registered_search_func: dict[str, SearchFunctionInfo] = {}


def search_function(*modes: str) -> Callable[[TSF], TSF]:
    def deco(func: TSF) -> TSF:
        info = SearchFunctionInfo(func)
        for mode in modes:
            registered_search_func[mode] = info
        return func

    return deco



import asyncio
import operator
import re
from collections.abc import Awaitable, Callable, Iterable
from contextlib import suppress
from difflib import SequenceMatcher
from functools import wraps
from io import BytesIO
from pathlib import Path
from typing import (
    TypeVar,
)
from typing_extensions import ParamSpec

import arrow
from cookit.loguru import logged_suppress
from httpx import URL, AsyncClient, HTTPStatusError, InvalidURL
from nonebot.matcher import current_bot, current_event, current_matcher
from nonebot_plugin_alconna.uniseg import Image as ImageSeg, UniMessage, image_fetch
from PicImageSearch.model.ehentai import EHentaiItem, EHentaiResponse
from PIL import Image
from pyquery import PyQuery
from tenacity import TryAgain, retry, stop_after_attempt, stop_after_delay

from .config import config
from .nhentai_model import NHentaiItem, NHentaiResponse

T = TypeVar("T")
P = ParamSpec("P")


DEFAULT_HEADERS = {
    "User-Agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/99.0.4844.82 Safari/537.36"
    ),
}


def post_image_process(file: bytes) -> bytes:
    im = Image.open(BytesIO(file))
    if (im.format == "WEBP") or getattr(im, "is_animated", False):
        with BytesIO() as output:
            im.save(output, "PNG")
            return output.getvalue()
    return file


@retry(stop=(stop_after_attempt(3) | stop_after_delay(30)), reraise=True)
async def get_image_bytes_by_url(url: str, cookies: str | None = None) -> bytes:
    _url = URL(url)
    referer = f"{_url.scheme}://{_url.host}/"
    headers = DEFAULT_HEADERS if _url.host.endswith("qpic.cn") else {"Referer": referer, **DEFAULT_HEADERS}
    async with AsyncClient(
        headers=headers,
        cookies=parse_cookies(cookies),
        proxy=config.proxy,
        follow_redirects=True,
    ) as session:
        resp = await session.get(url)
        if resp.status_code == 404:
            resp.raise_for_status()
            raise Exception  # NoReturn for sure, just make linter know
        if resp.status_code >= 400 or len(resp.content) == 0:
            raise TryAgain
        return post_image_process(resp.content)


async def get_image_from_seg(seg: ImageSeg) -> bytes:
    with suppress(ValueError):
        return seg.raw_bytes
    if seg.path:
        return Path(seg.path).read_bytes()
    if file := await image_fetch(
        current_event.get(),
        current_bot.get(),
        current_matcher.get().state,
        seg,
    ):
        return file
    raise ValueError("Cannot get image")


async def handle_img(
    url: str,
    hide_img: bool = config.hide_img,
    cookies: str | None = None,
) -> UniMessage:
    if not hide_img:
        with logged_suppress("Failed to get image", HTTPStatusError):
            return UniMessage.image(raw=await get_image_bytes_by_url(url, cookies))
    return UniMessage.text(f"预览图链接：{url}")


def handle_source(source: str) -> str:
    return (
        source.replace("www.pixiv.net/en/artworks", "www.pixiv.net/artworks")
        .replace(
            "www.pixiv.net/member_illust.php?mode=medium&illust_id=",
            "www.pixiv.net/artworks/",
        )
        .replace("http://", "https://")
    )


def parse_source(resp_text: str, host: str) -> str | None:
    if host in {"danbooru.donmai.us", "gelbooru.com"}:
        source = PyQuery(resp_text)(".image-container").attr("data-normalized-source")
        return str(source) if source else None

    if host in {"yande.re", "konachan.com"}:
        source = PyQuery(resp_text)("#post_source").attr("value")
        pool_text = PyQuery(resp_text)('a[href^="/pool/show/"]').text() or ""
        return str(source) or str(pool_text)

    return ""


async def get_source(url: str) -> str:
    if not url:
        return ""

    _url = get_valid_url(url)
    if not _url:
        return ""

    host = _url.host
    headers = None if host == "danbooru.donmai.us" else DEFAULT_HEADERS
    async with AsyncClient(
        headers=headers,
        proxy=config.proxy,
        follow_redirects=True,
    ) as session:
        resp = await session.get(url)
        if resp.status_code >= 400:
            return ""

        source = parse_source(resp.text, host)
        if source and get_valid_url(source):
            return handle_source(source)

    return source or ""


def confuse_url(url: str) -> str:
    return next(
        (
            url.replace("//", "// ").replace(host, host.replace(".", ". "))
            for host in config.to_confuse_urls
            if host in url
        ),
        url,
    )


async def shorten_url(url: str, force_shorten: bool = False) -> str:
    pid_search = re.compile(
        r"(?:pixiv.+(?:illust_id=|artworks/)|/img-original/img/(?:\d+/){6})(\d+)",
    )
    if pid_match := pid_search.search(url):
        return confuse_url(f"https://pixiv.net/i/{pid_match[1]}")

    uid_search = re.compile(r"pixiv.+(?:member\.php\?id=|users/)(\d+)")
    if uid_match := uid_search.search(url):
        return confuse_url(f"https://pixiv.net/u/{uid_match[1]}")

    host = URL(url).host
    if host == "danbooru.donmai.us":
        return confuse_url(url.replace("/post/show/", "/posts/"))

    if force_shorten or host in {
        "e-hentai.org",
        "exhentai.org",
        "graph.baidu.com",
        "nhentai.net",
        "www.google.com",
        "yandex.com",
    }:
        flag = len(url) > 1024
        async with AsyncClient(headers=DEFAULT_HEADERS) as session:
            if not flag:
                resp = await session.post("https://yww.uy/shorten", json={"url": url})
                if resp.status_code < 400:
                    return resp.json()["url"]
                flag = True
            if flag:
                resp = await session.post(
                    "https://www.shorturl.at/shortener.php",
                    data={"u": url},
                )
                if resp.status_code < 400:
                    final_url = PyQuery(resp.text)("#shortenurl").attr("value")
                    return f"https://{final_url}"

    return confuse_url(url)


def parse_cookies(cookies_str: str | None = None) -> dict[str, str]:
    cookies_dict: dict[str, str] = {}
    if cookies_str:
        for line in cookies_str.split(";"):
            key, value = line.strip().split("=", 1)
            cookies_dict[key] = value
    return cookies_dict


def async_lock(
    freq: float = 1,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Awaitable[T]]]:
    def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        lock = asyncio.Lock()
        last_call_time: arrow.Arrow | None = None

        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            nonlocal last_call_time
            async with lock:
                elapsed_time = arrow.now() - (last_call_time or arrow.now().shift(seconds=-freq))
                if elapsed_time.total_seconds() < freq:
                    await asyncio.sleep(freq - elapsed_time.total_seconds())
                result = await func(*args, **kwargs)
                last_call_time = arrow.now()
                return result

        return wrapper

    return decorator


def preprocess_search_query(query: str) -> str:
    query = re.sub(r"●|・|~|～|〜|、|×|:::|\s+-\s+|\[中国翻訳]", " ", query)
    # 去除独立的英文、日文、中文字符，但不去除带连字符的
    for i in [
        r"\b[A-Za-z]\b",
        r"\b[\u4e00-\u9fff]\b",
        r"\b[\u3040-\u309f\u30a0-\u30ff]\b",
    ]:
        query = re.sub(rf"(?<!-){i}(?!-)", "", query)

    return query.strip()


T_Item = TypeVar("T_Item", EHentaiItem, NHentaiItem)
T_Response = TypeVar("T_Response", EHentaiResponse, NHentaiResponse)


def filter_results_with_ratio(
    res: T_Response,
    title: str,
) -> list[T_Item]:
    raw_with_ratio = [(i, SequenceMatcher(lambda x: x == " ", title, i.title).ratio()) for i in res.raw]
    raw_with_ratio.sort(key=operator.itemgetter(1), reverse=True)

    filtered = [i[0] for i in raw_with_ratio if i[1] > 0.65]
    return filtered or [i[0] for i in raw_with_ratio]  # type: ignore


def get_valid_url(url: str) -> URL | None:
    with suppress(InvalidURL):
        url_obj = URL(url)
        if url_obj.host:
            return url_obj
    return None


def combine_message(
    msg_list: Iterable[UniMessage | str | None],
    join: str | None = "\n",
) -> UniMessage:
    msg = UniMessage()
    for i, it in enumerate(msg_list):
        if not it:
            continue
        if join and i != 0:
            msg += join
        msg += it
    return msg
